/*
 * Copyright (c) 2007, Thomas M�ller, xnos Internet Services (xnos.org)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *  - Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer.
 *  - Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *  - Neither the name of xnos nor the names of its contributors may be used 
 *    to endorse or promote products derived from this software without 
 *    specific prior written permission.
 *    
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, 
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 * Original author:  Thomas Muller, xnos Internet Services (xnos.org), 2007
 */ 

/*
 * Licensed Materials - Property of IBM
 *
 * trousers - An open source TCG Software Stack
 *
 * (C) Copyright International Business Machines Corp. 2004, 2005
 *
 */


#ifdef WIN32
#include <e_os.h>
#endif

#include <errno.h>
#include <string.h>


#include "trousers/tss.h"
#include "trousers_types.h"
#include "linux/tpm.h"
#include "tcslog.h"
#include "tddl.h"

// Need to link with Ws2_32.lib, Mswsock.lib, and Advapi32.lib
#define DEFAULT_BUFLEN 512

#ifndef TDDL_REMOTEPORT
#define TDDL_REMOTEPORT 27015
#endif

#ifndef TDDL_REMOTEHOST
#define TDDL_REMOTEHOST "130.192.1.249"
#endif

int fdsock = -1;
BYTE txBuffer[TDDL_TXBUF_SIZE];
int connection_ready = 0;

void hexdump(unsigned char *buf, int len) {
	while(--len >=0)
		printf("%02x", *buf++);
	printf("\n");
}


int
init_sock(void)
{
	TSS_RESULT result;
	struct hostent *hEnt = NULL;
	struct sockaddr_in addr;
	int portnum = TDDL_REMOTEPORT;
	
#ifdef WIN32
	WSADATA wsaData;
	struct addrinfo hints, *solved_addr = NULL;
	char port[10];
	
	// Initialize Winsock
	result = WSAStartup(MAKEWORD(2,2), &wsaData);
	if(result) {
		LogError("hostname %s does not resolve to a valid address.", TDDL_REMOTEHOST);	
		return -1;
	}
	   
	ZeroMemory( &hints, sizeof(hints) );
	hints.ai_family = PF_INET;
	hints.ai_socktype = SOCK_STREAM;
#endif
	fdsock = socket(PF_INET, SOCK_STREAM, 0);
	if (fdsock == -1) {
		LogError("socket: %s", strerror(errno));
		result = -1;
		goto err_exit;
	}

	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(portnum);

	LogDebug("Connecting to host %s.", TDDL_REMOTEHOST);

	/* try to resolve by hostname first */
	hEnt = gethostbyname(TDDL_REMOTEHOST);
	if (hEnt == NULL) {
#ifdef WIN32
		//itoa(get_port(), port, 10); 
		snprintf(port, sizeof port, "%d", portnum);
		
		result = getaddrinfo((char *)TDDL_REMOTEHOST, port, &hints, &solved_addr);		//htons?
		if(result != TSS_SUCCESS) {
			LogError("getaddrinfo failed: %d\n", result);
			result = -1;
			goto err_exit;
		}
		memcpy(&addr.sin_addr, solved_addr->ai_addr, 4);
#else
		/* if by hostname fails, try by dot notation */
		if (inet_aton((char *)TDDL_REMOTEHOST, &addr.sin_addr) == 0) {
			LogError("hostname %s does not resolve to a valid address.", TDDL_REMOTEHOST);
			result = -1;
			goto err_exit;
		}
#endif
	} else {
		memcpy(&addr.sin_addr, hEnt->h_addr_list[0], 4);
	}

	LogDebug("Connecting to %s", inet_ntoa(addr.sin_addr));

	if (connect(fdsock, (struct sockaddr *) &addr, sizeof (addr))) {
		LogError("connect: %s", strerror(errno));
		result = -1;
		goto err_exit;
	}

	return TSS_SUCCESS;

err_exit:
	close(fdsock);
#ifdef WIN32
	if(solved_addr)
	      freeaddrinfo(solved_addr);
	WSACleanup();
#endif
	return result;
}



TSS_RESULT
Tddli_Open()
{
	int rc;

	if (connection_ready) {
		LogDebug("attempted to re-open the TPM driver!");
		return -1;
	}

	rc = init_sock();
	if (rc) {
		LogError("Could not contact the TPM!");

		return -1;
	}
	connection_ready = 1;
	return TSS_SUCCESS;
}


TSS_RESULT
Tddli_Close()
{
	if (connection_ready == FALSE) {
		LogDebug("attempted to re-close the TPM driver!");
		return TDDLERR(TDDL_E_ALREADY_CLOSED);
	}

	close(fdsock);
#ifdef WIN32
	WSACleanup();
#endif
	connection_ready = 0;

	return TSS_SUCCESS;
}


int sock_write(const BYTE *buf, UINT32 buflen) {
    int iResult;

    iResult = send(fdsock, (const char *) buf, buflen, 0 );
    if (iResult <= 0) {
	printf("send failed: %s\n", strerror(errno));
	close(fdsock);
#ifdef WIN32
	WSACleanup();
#endif
	connection_ready = FALSE;
	return -1;
    }
//	printf("-> ");
//	hexdump((unsigned char *) buf, buflen);
    return iResult;
}


int sock_read(BYTE *buf, UINT32 buflen) {
    int iResult;

	iResult = recv(fdsock, (char *) buf, buflen, 0);
//	printf("<- ");
//	hexdump(buf, strlen((const char *) buf));

	return iResult;
}


TSS_RESULT
Tddli_TransmitData(BYTE * pTransmitBuf, UINT32 TransmitBufLen, BYTE * pReceiveBuf,
		   UINT32 * pReceiveBufLen)
{
	int sizeResult;

	if (TransmitBufLen > TDDL_TXBUF_SIZE) {
		LogError("buffer size handed to TDDL is too large! (%u bytes)", TransmitBufLen);
		return TDDLERR(TDDL_E_FAIL);
	}

	memcpy(txBuffer, pTransmitBuf, TransmitBufLen);
	LogDebug("Calling write to driver");

	if ((sizeResult = sock_write(txBuffer, TransmitBufLen)) == (int)TransmitBufLen)
		sizeResult = sock_read(txBuffer, TDDL_TXBUF_SIZE);

	if (sizeResult < 0) {
		LogError("read from device failed: %s", strerror(errno));
		return TDDLERR(TDDL_E_IOERROR);
	} else if (sizeResult == 0) {
		LogError("Zero bytes read from device");
		return TDDLERR(TDDL_E_IOERROR);
	}

	if ((unsigned)sizeResult > *pReceiveBufLen) {
		LogError("read %d bytes (only room for %d)", sizeResult, *pReceiveBufLen);
		return TDDLERR(TDDL_E_INSUFFICIENT_BUFFER);
	}

	*pReceiveBufLen = sizeResult;

	memcpy(pReceiveBuf, txBuffer, *pReceiveBufLen);
	return TSS_SUCCESS;
}


TSS_RESULT
Tddli_GetStatus(UINT32 ReqStatusType, UINT32 *pStatus)
{
	return TDDLERR(TSS_E_NOTIMPL);
}


TSS_RESULT
Tddli_SetCapability(UINT32 CapArea, UINT32 SubCap,
		    BYTE *pSetCapBuf, UINT32 SetCapBufLen)
{
	return TDDLERR(TSS_E_NOTIMPL);
}


TSS_RESULT
Tddli_GetCapability(UINT32 CapArea, UINT32 SubCap,
		    BYTE *pCapBuf, UINT32 *pCapBufLen)
{
	return TDDLERR(TSS_E_NOTIMPL);
}


TSS_RESULT
Tddli_Cancel(void)
{
	return TDDLERR(TSS_E_NOTIMPL);
}

